import QUB
from Buttons import *
from qubtree import Node
import math
import zlib
import struct
import os
from os.path import isfile

### -------------------------------------------------------------------------
### create a SWF movie from idealized data.
### Insert a frame (nClasses+1) on segment changes 
#
# IntToBin( i, iBits ) :
# SWFFB( nVal, iBits ) :
# BinToSWFBit( c ) :
# SWFRectangle(iBits, ix0, ix1, iy0, iy1 ) :
# SWFMatrix( iXScale, iYScale, iXSkew, iYSkew, iXTranslate, iYTranslate ) :
# SWFShowFrame(iCount) :
# SWFEndFile() :
# RGB(r,g,b) :
# RGBA(r,g,b) :
# SWFDefineBits2(id, iWid, iHgt, cPalette, cImage) :
# 
# SWFBitmapShapeWithStyle( iCharID, iX0, iX1, iY0, iY1 ) :
#   SWFFillStyleBitmap( iCharID ) :
#   SWFLineStyle( iPixelWidth, r,g,b ) :
#
# SWFPlaceObject2( iDepth, iCharID ) :
# SWFRemoveObject2( iDepth ) :
# SWFSetBGR( r,g,b ) :
# BMPLoad(cFile,iID) -> (iWid, iHgt, cImage)
# SWFFile( cRct, iFrameRate, iFrameCount , cData ) :

#-- convert int to binary string.  Ex : (25,6) -> '011001'
def IntToBin( i, iBits ) :
    i=int(i)
    iBits=int(iBits)
    c=''
    for iBit in range(iBits):
        if i & (2**(iBits-iBit-1)) :
            c=c+'1'
        else:
            c=c+'0'
    return c

#-- convert float to binary string
# Float is stored in 16+iBits bits - first iBits are integer, next 16 are fraction of 2^16
def SWFFB( nVal, iBits ) :
    return IntToBin(int(nVal),iBits) + IntToBin(int((nVal-math.floor(nVal))*(2**16)),16)
    
#-- Convert binary string to ascii.  Ex : '01000001' -> 'A'
def BinToSWFBit( c ) :
    cc = '' 
    c = c + '00000000'
    while len(c)>8 :
        cc = cc + chr(int(c[0])*128 +int(c[1])*64 +int(c[2])*32 +int(c[3])*16 +int(c[4])*8 +int(c[5])*4 +int(c[6])*2 +int(c[7]))
        c = c[8:]        
    return cc

#-- Create SWF rectagle string : 1st 5 bits are number of bits for each subsequent
#-- value.  Followed by x0,x1,y0,y1 and padded with 0's to byte length.
def SWFRectangle(iBits, ix0, ix1, iy0, iy1 ) :
    if iBits == 0 :
        iBits=math.ceil(math.log(max(ix0,ix1,iy0,iy1),2))+1
    return BinToSWFBit(IntToBin(iBits,5)+IntToBin(ix0,iBits)+IntToBin(ix1,iBits)+IntToBin(iy0,iBits)+IntToBin(iy1,iBits))

#-- 
def SWFMatrix( nXScale, nYScale, nXSkew, nYSkew, iXTranslate, iYTranslate ) :
    c = ''

    if nXScale==-1 :
        c='0'
    else :
        iBits=int(math.ceil(math.log(max(nXScale*20.0,nYScale*20.0),2)))+1
        c = '1' + IntToBin(iBits+16,5) + SWFFB(nXScale*20.0,iBits) +SWFFB(nYScale*20.0,iBits)

    if nXSkew==-1 :
        c=c+'0'
    else :
        iBits=int(math.ceil(math.log(max(iXSkew,iYSkew),2)))+1
        c = c + '1' +IntToBin(iBits+16,5) +SWFFB(nXSkew,iBits) +SWFFB(nYSkew,iBits)

    if iXTranslate==-1 :
        c=c+'00000'
    else :
        iBits=int(math.ceil(math.log(max(iXTranslate*20,iYTranslate*20),2)))+1
        c=c+IntToBin(iBits,5)+IntToBin(int(iXTranslate*20),iBits)+IntToBin(int(iYTranslate*20),iBits)
    return BinToSWFBit(c)        

#-- Create string of <iCount> SWF file Show Frame commands
def SWFShowFrame(iCount) :
    return (chr(64)+chr(0))*int(iCount)

#-- Required at end of SWF file
def SWFEndFile() :
    return chr(0)+chr(0)

#-- Convert (R,G,B) to 4 byte string (alpha=255)
def RGB(r,g,b) :
    return struct.pack('l',b*0x10000+g*0x100+r)[0:3]
def RGBA(r,g,b) :
    return struct.pack('l',0xFF000000+r*0x10000+g*0x100+b)

#-- Create SWF lossless Bitmap type 3 - 256 color bitmap index
def SWFDefineBits2(id, iWid, iHgt, cPalette, cImage) :
    cComp = zlib.compress(cPalette+cImage,9)
    return chr(63)+chr(9) +struct.pack('l',8+len(cComp)) +struct.pack('h',id) +chr(3) +struct.pack('h',iWid) +struct.pack('h',iHgt) + chr(len(cPalette)/4-1) + cComp

#-- Create SWF lossless Bitmap type 5 - RGBA array (Alpha   .Blue.Green.Red)
def SWFDefineBits2a(id, iWid, iHgt, cImage) :
    cComp = zlib.compress(cImage,9)
    return chr(63)+chr(9) +struct.pack('l',7+len(cComp)) +struct.pack('h',id) +chr(5) +struct.pack('h',iWid) +struct.pack('h',iHgt) + cComp

#-- Bitmap character placement
def SWFFillStyleBitmap( iCharID, iDX, iDY ) :
    return chr(0x41) + struct.pack('h',iCharID) + SWFMatrix(1.0,1.0,-1,-1,iDX,iDY)
def SWFLineStyle( iPixelWidth, r,g,b ) :
    return struct.pack('h',iPixelWidth*20) + RGB(r,g,b)
def SWFBitmapShapeWithStyle( iCharID, iX0, iX1, iY0, iY1 ) :
    dY=iY1-iY0
    dX=iX1-iX0

    #----- SHAPEWITHSTYLE tag contents
    # Fill Styles
    c = chr(2) + SWFFillStyleBitmap( 0x7FFF, iX0, iY0 ) + SWFFillStyleBitmap( iCharID, iX0, iY0 )
    # line Styles
    c = c + chr(0)
    # Fill, line index bits = 2,0
    c = c + chr(0x20)
    # Style change - NEWSTYLE + EDGE + EDGE + EDGE + EDGE + END
    # type=0 newstyle=0 linestyle=0 fill1=1 fill0=0 move=1 + movesize[5] + dx[] + dy[] + Fill1=2
    iBits=math.ceil(math.log(20*max(iX1,iY0),2))+1
    cc = '000101' + IntToBin(iBits,5) +IntToBin(iX1*20,iBits) + IntToBin(iY0*20,iBits) +'10'
    # y=y+dy : type=1(edge) straight=1 bits[4](add+2) general=0(V or H) 1(=Ver) DY
    iBits=math.ceil(math.log(20*dY,2))+1
    cc = cc + '11'+IntToBin(iBits-2,4)+'01'+IntToBin(20*dY,iBits)
    # x=x-dx : type=1(edge) straight=1 bits[4](add+2) general=0(V or H) 0(=Hor) DX
    iBits=math.ceil(math.log(20*dY,2))+1
    cc = cc + '11'+IntToBin(iBits-2,4)+'00'+IntToBin(-20*dX,iBits)
    # y=y-dy 
    iBits=math.ceil(math.log(20*dY,2))+1
    cc = cc + '11'+IntToBin(iBits-2,4)+'01'+IntToBin(-20*dY,iBits)
    # x=x+dx : type=1(edge) straight=1 bits[4](add+2) general=0(V or H) 0(=Hor) DX
    iBits=math.ceil(math.log(20*dY,2))+1
    cc = cc + '11'+IntToBin(iBits-2,4)+'00'+IntToBin(20*dX,iBits)
    # End Shape Record
    cc = cc + '000000'
    return c + BinToSWFBit( cc )

def SWFRectangleShapeWithStyle( iCharID, iX0, iX1, iY0, iY1, iRed, iGrn, iBlu ) :
    dY=iY1-iY0
    dX=iX1-iX0

    #----- SHAPEWITHSTYLE tag contents
    # 0 Fill Styles, 1 line syle = 20/20 pixels blue, 0 fill index bits, 1 line index bit
    c = chr(0) + chr(1) + SWFLineStyle( 1, iRed, iGrn, iBlu ) + chr(1)

    # Style change - NEW-LINE-STYLE&MOVE  + EDGE + EDGE + EDGE + EDGE + END
    # type=0 newstyle=0 linestyle=1 fill1=0 fill0=0 move=1 + movesize[5] + dx[] + dy[]
    iBits=math.ceil(math.log(20*max(iX1,iY0),2))+1
    cc = '001001' + IntToBin(iBits,5) + IntToBin(iX1*20,iBits) + IntToBin(iY0*20,iBits) +'1'
    # y=y+dy : type=1(edge) straight=1 bits[4](add+2) general=0(V or H) 1(=Ver) DY
    iBits=math.ceil(math.log(20*dY,2))+1
    cc = cc + '11'+IntToBin(iBits-2,4)+'01'+IntToBin(20*dY,iBits)
    # x=x-dx : type=1(edge) straight=1 bits[4](add+2) general=0(V or H) 0(=Hor) DX
    iBits=math.ceil(math.log(20*dX,2))+1
    cc = cc + '11'+IntToBin(iBits-2,4)+'00'+IntToBin(-20*dX,iBits)
    # y=y-dy 
    iBits=math.ceil(math.log(20*dY,2))+1
    cc = cc + '11'+IntToBin(iBits-2,4)+'01'+IntToBin(-20*dY,iBits)
    # x=x+dx : type=1(edge) straight=1 bits[4](add+2) general=0(V or H) 0(=Hor) DX
    iBits=math.ceil(math.log(20*dX,2))+1
    cc = cc + '11'+IntToBin(iBits-2,4)+'00'+IntToBin(20*dX,iBits)
    # End Shape Record
    cc = cc + '000000'
    return c + BinToSWFBit( cc )

#--
def SWFBitmapPlace( iShapeID, iCharID, iX0, iX1, iY0, iY1 ) :
    cc = SWFBitmapShapeWithStyle( iCharID, iX0, iX1, iY0, iY1 )
    cRct = SWFRectangle(0, int(20*iX0), int(20*iX1), int(20*iY0), int(20*iY1) )
    ilen = len(cc)+len(cRct)+2
    return chr(0xBF)+chr(0) +struct.pack('l',ilen) +struct.pack('h',iShapeID) + cRct + cc

def SWFRectanglePlace( iShapeID, iCharID, iX0, iX1, iY0, iY1, iRed, iGrn, iBlu ) :
    cc = SWFRectangleShapeWithStyle( iCharID, iX0, iX1, iY0, iY1, iRed, iGrn, iBlu )
    cRct = SWFRectangle(0, int(20*iX0), int(20*iX1), int(20*iY0), int(20*iY1) )
    ilen = len(cc)+len(cRct)+2
    return chr(0xBF)+chr(0) +struct.pack('l',ilen) +struct.pack('h',iShapeID) + cRct + cc

#-- Place a previously defined shape object 
def SWFPlaceObject2( iDepth, iCharID ) :
    # change flags 6->2 and len 6->5 and remove matrix at end.
    return chr(0x86)+chr(6) +chr(6) +struct.pack('h',iDepth) +struct.pack('h',iCharID)+chr(0)

#-- Move a previously placed shape object 
def SWFMoveObject2( iDepth, iDX, iDY ) :
    #cRct must be 60 bytes max ( which it always is)
    cMatrix = SWFMatrix(-1,-1,-1,-1,iDX,iDY)
    #SWFRectangle(0, int(20*iX0), int(20*iX1), int(20*iY0), int(20*iY1) )
    return chr(128+(3+len(cMatrix)))+chr(6) +chr(5) + struct.pack('h',iDepth) + cMatrix

#-- Remove object at specified depth 
def SWFRemoveObject2( iDepth ) :
    return chr(2)+chr(7)+struct.pack('h',iDepth)

#-- Set background tag (id=9)
def SWFSetBGR( r,g,b ) : 
    return chr(0x43)+chr(2)+chr(r)+chr(g)+chr(b)

#-- Construct file header
def SWFFile( cOutFile, cRct, iFrameRate, iFrameCount , cData ) :
    f = open(cOutFile, 'wb')
    f.write('FWS'+chr(5)+struct.pack('l',12+len(cRct)+len(cData)) +cRct)
    f.write(chr(0)+chr(iFrameRate) + struct.pack('h',iFrameCount))
    f.write(cData)
    f.close()
    return 

#---------------------------------------------------------------------------------
def BMPLoad(cFile, iID) :
    f = file(cFile,'rb')
    f.seek(10)
    iPos=struct.unpack('l',f.read(4))[0]
    f.seek(18)
    iWid=struct.unpack('l',f.read(4))[0]
    iHgt=struct.unpack('l',f.read(4))[0]
    f.seek(28)
    iBitsPerPixel=struct.unpack('h',f.read(2))[0]
    iBytesPerPixel=iBitsPerPixel/8
    iBytesPerRow=iBytesPerPixel*iWid
    iHgt=min(800,iHgt)
    cImage=''

    f.seek(iPos)
    for iRow in range(iHgt) :
        cc=''
        for iCol in range(iWid) :
            cRGB=f.read(iBytesPerPixel)
            if iCol < 800 :
                cc=cc+chr(255)+cRGB[2]+cRGB[1]+cRGB[0]
        if ((iWid*iBytesPerPixel)%4)<>0:
            f.read(4-((iWid*iBytesPerPixel)%4))
        cImage=cc+cImage
    iWid = min(800,iWid)
    f.close()
    print 'Loaded',iHgt,iWid,"Name:"+cFile
    return (iHgt,iWid,SWFDefineBits2a(iID, iWid, iHgt, cImage ))

### -------------------------------------------------------------------------
### create a PYMOL script from idealized data.
### Insert a frame (nClasses+1) on segment changes 
def TAIDLScript(props) : 
    #----- Properties
    iFramesPerSec = props['framespersec']
    iMinFrames = props['minframes']
    iMaxFrames = props['maxframes'] 
    iBlankSep = props['blanksep']
    iBlankFrame = props['blankframe']

    #----- Scan data segments 
    cScript = ''
    cLine='mset '
    xidl=QUB.getChosenIdealization()
    xseg=xidl.find('DataSet').find('Segment')
    while not xseg.isNull:
        xcla=xseg.find('Classes')
        xdur=xseg.find('Durations')
        for iSeg in range(xseg.find('DwellCount').data[0]) : 
            iFrames = int(xdur.data[iSeg]*iFramesPerSec/1000.0)
            iFrames = min(max(iFrames,iMinFrames),iMaxFrames)
            if iFrames > 0 :
                cLine = cLine + str(xcla.data[iSeg]+1)+' x'+str(iFrames)+'  '
            if len(cLine) > 60 : 
                cScript = cScript + cLine + '\n'
                cLine = 'madd  '
                
        cScript = cScript + cLine + str(iBlankFrame)+' x'+str(iFrames)+'\n'
        cLine = 'madd '
        xseg = xseg.next('Segment')
    QUB.Report( cScript )

### -------------------------------------------------------------------------
AddTaskButton( 'Tony Idl PYMOL Script', # button text
               TAIDLScript,  # function
               'PYMO',       # short text for small buttons
               0,            # show settings dialog on left-click?
               1,            # show settings dialog on right-click?
               "Pymol script from Idealized",   # tool-tip
               "Pymol script from Idealized",   # description in "Edit Buttons"
               "Pymol script from Idealized:", # caption for the settings dialog
               [TextField("framespersec", "PYMOL frames per seconds of idealization :", "50", acceptIntGreaterThan(1)),
               TextField("minframes",     "Minimum frames for short events:", "0", acceptIntGreaterThan(-1)),
               TextField("maxframes",     "Maximum frames per event :", "1000", acceptIntGreaterThan(-1)),
               TextField("blankframe",      "Frame number for blanks:", "99", acceptIntGreaterThan(-1)),
               TextField("blanksep",      "Blank frames between segments:", "15", acceptIntGreaterThan(-1))
               ],
               'Preprocessing' )

### -------------------------------------------------------------------------
### create a SWF movie from idealized data.
### Insert a frame (nClasses+1) on segment changes
### mage ID #'s assume max of 100 frames 
def TASWF(props) : 
    #----- Properties
    iFramesPerSec = 12
    nSecPerSec = props['secdiv']
    iMinFrames = props['minframes']
    iMaxFrames = props['maxframes']
    cFirstImage = props['FirstImage']
    cOutFile = props['OutFile']
    #iBlankSep = props['blanksep']
    #iBlankFrame = props['blankframe']
    xIndic=props['xIndic']
    yIndic=props['yIndic']
    widIndic=props['widIndic']
    hgtIndic=props['hgtIndic']

    #----- Scan data segments
    aIDL = []
    iTotalFrames=0
    iMaxClass=1
    xidl=QUB.getChosenIdealization()
    xseg=xidl.find('DataSet').find('Segment')
    nTimeElapsed = 0.0
    while not xseg.isNull:
        xcla=xseg.find('Classes')
        xdur=xseg.find('Durations')
        for iSeg in range(xseg.find('DwellCount').data[0]) :
            iClass = xcla.data[iSeg]
            iFrames = min(iMaxFrames, max(iMinFrames, int(0.5+nSecPerSec * (xdur.data[iSeg]/1000.0) * iFramesPerSec) ))
            nTimeElapsed=nTimeElapsed+xdur.data[iSeg]
            if iFrames > 0 : 
                iMaxClass = max(iMaxClass,iClass)
                aIDL.append( [iClass,iFrames,nTimeElapsed] )
                iTotalFrames=iTotalFrames+iFrames
        xseg = xseg.next('Segment')
    nTimeTotal = nTimeElapsed

    #----- Image 'character' definitions
    iMaxHgt=1
    iMaxWid=1
    cSWF = SWFSetBGR(64,64,64)
    aImageCount=[]

    for i in range(0, iMaxClass+1) :
        cCharImage = cFirstImage[0:-4]+'a.bmp'
        if isfile(cCharImage) :
            j=0
            while isfile(cCharImage) :
                j=j+1
                oImage = BMPLoad( cCharImage, i*100+100+j)
                iMaxHgt = max(iMaxHgt,oImage[0])
                iMaxWid = max(iMaxWid,oImage[1])
                cSWF = cSWF + oImage[2]
                cCharImage = cCharImage[0:-5]+chr(65+j)+'.bmp'
            aImageCount.append( j )
        else :
            oImage = BMPLoad( cFirstImage, i+1)
            iMaxHgt = max(iMaxHgt,oImage[0])
            iMaxWid = max(iMaxWid,oImage[1])
            cSWF = cSWF + oImage[2]
            aImageCount.append( 0 )
        if cFirstImage[-5]=='9' : 
            cFirstImage = cFirstImage[0:-7]+str(int(cFirstImage[-6:-4])+1)+'.bmp'
        else :
            cFirstImage = cFirstImage[0:-5]+str(int(cFirstImage[-5])+1)+'.bmp'

    #----- Image Placements
    for i in range(0, iMaxClass+1) :
        if aImageCount[i]==0 :
            cSWF = cSWF + SWFBitmapPlace(iMaxClass+2+i, i+1, 1.0, iMaxWid, 1.0, iMaxHgt )
        else :
            for j in range(0, aImageCount[i]) :
                cSWF = cSWF + SWFBitmapPlace(i*100+151+j, i*100+101+j, 1.0, iMaxWid, 1.0, iMaxHgt )
                # same ID's ... ? 

    #----- Timing Rectangle
    if xIndic >= 0 : 
        #cSWF = cSWF + SWFRectanglePlace(91, 90, xIndic, xIndic+iMaxWid-10, 1, 12, 128,128,128 )
        #cSWF = cSWF + SWFPlaceObject2(10, 121 )
        cSWF = cSWF + SWFRectanglePlace(93, 92,  xIndic, xIndic+1, yIndic, yIndic+hgtIndic, 0,0,0 )
        cSWF = cSWF + SWFPlaceObject2(11, 93 )

    #----- Movie frames
    print 'Frames=',iTotalFrames
    iFrames = 0
    iCurrPos=0
    nTimeElapsed=0
    for i in range( len(aIDL) ) :
        #----- Covert Idealization to ShowFrames sequence
        if aImageCount[aIDL[i][0]]==0 : 
            cSWF = cSWF + SWFPlaceObject2(1,iMaxClass+2+aIDL[i][0])
            for j in range(0, aIDL[i][1] ) :
                cSWF = cSWF + SWFShowFrame(1)
                #----- Update Horizontal bar indicator
                iPos = (nTimeElapsed+(aIDL[i][2]-nTimeElapsed)*j/aIDL[i][1])/nTimeTotal*widIndic
                if (xIndic>=0) and (iPos>iCurrPos) :
                    cSWF = cSWF + SWFMoveObject2( 11, iPos, 0)
                    iCurrPos = iPos
            cSWF = cSWF + SWFRemoveObject2(1)
        else :
            for j in range(0, aIDL[i][1] ) :
                cSWF = cSWF + SWFPlaceObject2(1,aIDL[i][0]*100+151+(j%aImageCount[aIDL[i][0]])) + SWFShowFrame(1) + SWFRemoveObject2(1)            
                #----- Update Horizontal bar indicator 
                iPos = (nTimeElapsed+(aIDL[i][2]-nTimeElapsed)*j/aIDL[i][1])/nTimeTotal*widIndic
                if (xIndic>=0) and (iPos>iCurrPos) :
                    cSWF = cSWF + SWFMoveObject2( 11, iPos, 0)
                    iCurrPos = iPos
        iFrames = iFrames + aIDL[i][1]
        nTimeElapsed=aIDL[i][2]
    cSWF = cSWF + SWFEndFile()

    cRct = SWFRectangle(0,0,20*iMaxWid,0,20*iMaxHgt)      # bits,x0,x1,y0,y1

    SWFFile(cOutFile,cRct,12,iFrames,cSWF)
    print 'SWF output complete'
    os.startfile(cOutFile)


### -------------------------------------------------------------------------
AddTaskButton( 'Tony SWF export', # button text
                TASWF,  # function
                'SWF',       # short text for small buttons
                0,            # show settings dialog on left-click?
                1,            # show settings dialog on right-click?
                "SWF from Idealized",   # tool-tip
                "SWF from Idealized",   # description in "Edit Buttons"
                "SWF from Idealized:", # caption for the settings dialog
                [TextField("secdiv",    "SWF secs per sec of idl :", "0.1", acceptFloat),
                TextField("minframes",  "Min frames for short events:",         "0", acceptIntGreaterThan(-1)),
                TextField("maxframes",  "Max frames per event :",               "1000", acceptIntGreaterThan(-1)),
                TextField("FirstImage", "1st image (must end with 1.bmp):",       "C:\\Temp\\Image0001.bmp", acceptString),
                TextField("OutFile", "Output file name :",                          "C:\\Temp\\Test.SWF", acceptString),
                TextField("xIndic",   "Indicator X pos :(-1=none)", "0", acceptIntGreaterThan(-2)),
                TextField("yIndic",   "Indicator Y pos :", "0", acceptIntGreaterThan(-2)),
                TextField("widIndic", "Indicator Width :", "400", acceptIntGreaterThan(-2)),
                TextField("hgtIndic", "Indicator Height:", "50", acceptIntGreaterThan(-2))
                ],
                'Preprocessing' )
                #TextField("blanksep",   "Frames between segments:",           "15", acceptIntGreaterThan(-1)),
                #TextField("BlankImage", "Image between segments:", "", acceptString), 



